import datetime
from typing import Any, Literal, Optional
import requests
from pydantic import BaseModel
from talc.synthetic import (
    Document,
    QuestionGenerationConfig,
)
from talc.grading import GradingSet
from requests.exceptions import HTTPError


class TestCase(BaseModel):
    """Represents a single test case that should be run."""

    id: str
    question: str
    expected_response: str | None


class TestCaseWithRubric(BaseModel):
    """Represents a single test case that should be run."""

    id: str | None
    question: str
    expected_response: str
    scenario_data: dict[str, object]
    source_content: list[str]


class Grade(BaseModel):
    """The output of a single grader on a test case."""

    reason: str | None
    score: float
    grader: str


class GradedTestCase(BaseModel):
    """Represents a single test case that has been graded."""

    id: str
    question: str
    response: str
    expected_response: str
    grading_complete: bool
    grades: list[Grade]


class RunInfo(BaseModel):
    """Represents a test run."""

    id: str
    completion_progress: float
    grades: dict[str, float]
    test_cases: list[GradedTestCase] | None


class Dataset(BaseModel):
    """Represents a dataset."""

    id: str
    friendly_name: str | None


class DatasetCreationJob(BaseModel):
    """Represents the status for a long-running job that creates a dataset in the DB."""

    id: str
    dataset_id: str
    last_updated: datetime.datetime
    status: str
    error: str | None
    diagnostics: Any
    name: Optional[str] = None


class APIGenerationArgs(BaseModel):
    """Arguments for creating or getting an API Key."""

    access_token: str


class KnowledgeBase(BaseModel):
    """A collection of documents, with relationships between them."""

    id: Optional[str] = None
    friendly_name: Optional[str] = None
    documents: list[str] = []  # List of document IDs
    errors: list[str] | None = None

    # FastAPI returns exception details in field .detail, but
    # Python's standard Response does not check this field.


def handle_fastAPI_errors(response):
    if response.status_code > 400:
        if response.headers["content-type"] == "application/json":
            raise HTTPError(response.json()["detail"], response=response)
        else:
            raise HTTPError(response.text, response=response)


class TalcClient:
    """Client for interacting with the Talc evals server."""

    def __init__(self, api_key: str, url: str, run_id: str | None = None) -> None:
        self.url: str = url
        self.run_id: str | None = run_id
        self.api_key: str = api_key

    def start_run(self) -> RunInfo:
        """Create a new test run."""

        response = requests.post(
            f"{self.url}/run/create", headers={"X-TALC-API": self.api_key}
        )
        handle_fastAPI_errors(response)

        info = RunInfo(**response.json())
        self.run_id = info.id
        return info

    def get_dataset(self, id: str) -> list[TestCaseWithRubric]:
        """Get all test cases in a single dataset from the server."""

        response = requests.get(
            f"{self.url}/dataset/{id}", headers={"X-TALC-API": self.api_key}
        )
        handle_fastAPI_errors(response)

        return [TestCaseWithRubric(**case) for case in response.json()]

    def list_kbs(self) -> list[KnowledgeBase]:
        """Get all datasets from the server."""

        response = requests.get(
            f"{self.url}/kb/list", headers={"X-TALC-API": self.api_key}
        )
        handle_fastAPI_errors(response)

        return [KnowledgeBase(**kb) for kb in response.json()]

    def submit_responses(self, to_grade: GradingSet) -> None:
        """Submit responses generated by the system for grading."""
        if self.run_id is None:
            raise RuntimeError("No run started. Remember to call start_run() first.")

        data = to_grade.dict()

        response = requests.post(
            f"{self.url}/run/{self.run_id}/submit",
            json=data,
            headers={"X-TALC-API": self.api_key},
        )
        handle_fastAPI_errors(response)

    def get_results(self) -> RunInfo:
        """Get graded results of a run."""
        if self.run_id is None:
            raise RuntimeError("No run started. Remember to call start_run() first.")

        response = requests.get(
            f"{self.url}/run/{self.run_id}/results",
            headers={"X-TALC-API": self.api_key},
        )
        handle_fastAPI_errors(response)

        return RunInfo(**response.json())

    def upload_dataset(
        self, friendly_name: str, dataset: list[TestCaseWithRubric]
    ) -> Dataset:
        """Upload a new dataset and save it to the database."""

        data = [x.dict() for x in dataset if x.id != "id"]

        response = requests.post(
            f"{self.url}/dataset/create/{friendly_name}",
            json=data,
            headers={"X-TALC-API": self.api_key},
        )
        handle_fastAPI_errors(response)

        return Dataset(**response.json())

    def start_generate_dataset(
        self,
        config: QuestionGenerationConfig,
    ) -> DatasetCreationJob:
        response = requests.post(
            f"{self.url}/dataset/generate/",
            json=config.dict(),
            headers={"X-TALC-API": self.api_key},
        )
        handle_fastAPI_errors(response)

        return DatasetCreationJob(**response.json())

    def get_dataset_creation_job(self, id: str) -> DatasetCreationJob:
        """Get the status of a dataset creation job."""
        response = requests.get(
            f"{self.url}/dataset/generate/{id}",
            headers={"X-TALC-API": self.api_key},
            timeout=10,
        )
        handle_fastAPI_errors(response)

        return DatasetCreationJob(**response.json())

    def get_dataset_status(self, id: str) -> DatasetCreationJob:
        """Get the status of a dataset creation job using the dataset ID."""
        response = requests.get(
            f"{self.url}/dataset/jobstatus/{id}",
            headers={"X-TALC-API": self.api_key},
            timeout=10,
        )
        response.raise_for_status()

        return DatasetCreationJob(**response.json())

    def get_recent_datasets(self) -> list[DatasetCreationJob]:
        """Get the status of the most recent dataset creation jobs."""
        response = requests.get(
            f"{self.url}/dataset/recent",
            headers={"X-TALC-API": self.api_key},
            timeout=10,
        )
        response.raise_for_status()

        return [DatasetCreationJob(**job) for job in response.json()]

    def upload_knowledge_base_asbytes(
        self,
        kb: list[Document],
        name: str,
        verify: bool = False,
        auto_link: bool = False,
    ) -> KnowledgeBase:
        """Upload a knowledge base to the server. The content of each document is uploaded as raw bytes."""
        if (
            isinstance(kb, list)
            and isinstance(kb[0], Document)
            and isinstance(kb[0].content, bytes)
        ):
            files = [
                ("files", (doc.filepath, doc.content, doc.content_type))
                for doc in kb
                if doc.content is not None
            ]
        else:
            raise ValueError(
                "Invalid knowledge base format. Only a list of non-null Documents with byte content supported."
            )

        response = requests.post(
            f"{self.url}/kb/create/{name}",
            params={"verify": verify, "auto_link": auto_link},
            files=files,  # type: ignore
            headers={"X-TALC-API": self.api_key},
        )
        handle_fastAPI_errors(response)
        return KnowledgeBase(**response.json())

    def create_analytics_event(
        self,
        event_name: str,
        event_data: dict[str, Any],
    ) -> None:
        """Create an analytics event."""
        response = requests.post(
            f"{self.url}/analytics",
            json={
                "event_name": event_name,
                "event_data": event_data,
            },
            headers={"X-TALC-API": self.api_key},
        )
        response.raise_for_status()

    # def get_knowledge_base(self, knowledgebase_id: str) -> list[DownloadDocument]:
    #     """Get a knowledge base from the server."""
    #     response = requests.get(
    #         f"{self.url}/kb/{knowledgebase_id}",
    #         headers={"X-TALC-API": self.api_key},
    #     )
    #     handle_fastAPI_errors(response)



import csv as csvmod
import datetime
import os
import random
from textwrap import indent
import time
from math import ceil

import click
from termcolor import colored

from talc.talc_client import (
    DatasetCreationJob,
    TalcClient,
    TestCaseWithRubric,
)
from talc.grading import (
    GradingSet,
    TestCaseResponse,
    ScoringRule,
    GradingPipeline,
)
from talc.synthetic import QuestionGenerationConfig, Document
from talc.utils.cli.io import print_single_case, get_failures_and_warnings, Web
from talc.utils.grading.defaults import DEFAULT_GRADER_CONFIG
from talc.utils.synthetic.defaults import (
    DEFAULT_HALLUCINATION_ARGS,
    DEFAULT_SIMPLE_ARGS,
    DEFAULT_MID_SIMPLE_ARGS,
)

talc_api_key: str | None = os.getenv("TALC_API_KEY", None)
talc_base_url: str = os.getenv("TALC_BASE_URL", "https://api.talcapi.com")


def poll_results(
    client: TalcClient,
    print_passed: bool,
    out_file: str | None,
    scoring_rules: list[ScoringRule],
):
    while True:
        run_info = client.get_results()
        print(f"Progress: {run_info.completion_progress * 100}%")
        if run_info.completion_progress >= 1:
            print("\n============================================================")
            print("Grading complete.")
            for key in run_info.grades.keys():
                print(f"{key}: {run_info.grades[key] * 100}%")

            print("============================================================\n")

            assert run_info.test_cases is not None

            failures, warnings = get_failures_and_warnings(
                run_info.test_cases, scoring_rules
            )

            passed = [
                case
                for case in run_info.test_cases
                if case not in failures and case not in warnings
            ]

            if out_file is not None:
                with open(
                    out_file, "w", newline="", encoding="utf-8", errors="replace"
                ) as csvfile:
                    writer = csvmod.writer(csvfile)
                    writer.writerow(
                        [
                            "id",
                            "question",
                            "expected_answer",
                            "result",
                            "failure",
                            "warning",
                            "grades",
                        ]
                    )
                    for case in run_info.test_cases:
                        writer.writerow(
                            [
                                case.id,
                                case.question,
                                case.expected_response,
                                case.response,
                                True if case in failures else False,
                                True if case in warnings else False,
                                case.grades,
                            ]
                        )

            if print_passed and len(passed) > 0:
                print("=================PASSED=================")
                for case in passed:
                    print_single_case(case, "green")
            if len(warnings) > 0:
                print("=================WARNINGS=================")
                for warning in warnings:
                    print_single_case(warning, "yellow")
            if len(failures) > 0:
                print("=================FAILURES=================")
                for failure in failures:
                    print_single_case(failure, "red")
            if len(failures) > 0:
                return 1
            else:
                return 0

        time.sleep(2)


@click.group()
def evals():
    pass


@evals.command()
@click.option(
    "--api_key",
    default=talc_api_key,
    help="Talc API key. Can also be specified with the TALC_API_KEY environment variable.",
)
@click.option(
    "--in_file",
    prompt="Results CSV path",
    help="Path to CSV file to load results from. Columns should be 'id', 'result'. Header row is required and case sensitive.",
)
@click.option("--print_passed", is_flag=True, help="Print the passed test cases.")
@click.option("--out_file", default=None, help="Optionally save output as CSV.")
@click.option(
    "--config",
    default="all_passed",
    help="The name of the grader configuration to use.",
)
def eval(
    api_key: str | None,
    in_file: str,
    print_passed: bool,
    out_file: str,
    config: str,
):
    """Grade a test run from a CSV. Columns 'id' and 'result' are required, other columns will be ignored. CSV should have a header row."""

    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    client = TalcClient(api_key, talc_base_url)

    with open(in_file, newline="", encoding="utf-8", errors="replace") as csvfile:
        reader = csvmod.DictReader(csvfile)

        results = [
            TestCaseResponse(id=row["id"], response=row["result"]) for row in reader
        ]

    if config.endswith(".json"):
        loaded_config = GradingPipeline.parse_file(config)
    else:
        loaded_config = DEFAULT_GRADER_CONFIG

    grading_set = GradingSet(
        responses=results,
        pipeline=loaded_config,  # type: ignore
    )
    print("Grading " + str(len(results)) + " test cases.")
    run_info = client.start_run()
    client.submit_responses(grading_set)
    print("Submitted results. Beginning grading for run ID: " + run_info.id)

    res = poll_results(client, print_passed, out_file, loaded_config.scoring_rules)
    exit(res)


@evals.command()
@click.option(
    "--api_key",
    default=talc_api_key,
    help="Talc API key. Can also be specified with the TALC_API_KEY environment variable.",
)
@click.option("--run_id", prompt="Run ID", help="The ID of the run to get results for.")
@click.option("--print_passed", is_flag=True, help="Print the passed test cases.")
def get_results(api_key: str | None, run_id: str, print_passed: bool):
    """Get the grades for a test run."""

    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    client = TalcClient(api_key, talc_base_url, run_id)

    res = poll_results(client, print_passed, None, DEFAULT_GRADER_CONFIG.scoring_rules)

    exit(res)


@evals.command()
@click.option(
    "--api_key",
    default=talc_api_key,
    help="Talc API key. Can also be specified with the TALC_API_KEY environment variable.",
)
@click.option(
    "--csv",
    prompt="CSV path",
    help="Path to CSV file to upload. Column headers should be 'question' and 'expected_answer'. An optional source_text column can be included as well.",
)
@click.option("--name", prompt="Dataset name", help="The name of the dataset.")
def upload_manual_testset(api_key: str | None, csv: str, name: str):
    """Add a new dataset to the server."""

    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    client = TalcClient(api_key, talc_base_url)

    with open(csv, newline="", encoding="utf-8", errors="replace") as csvfile:
        reader = csvmod.DictReader(csvfile)

        try:
            dataset = [
                TestCaseWithRubric(
                    question=row["question"],
                    expected_response=row["expected_answer"],
                    scenario_data={},
                    id=None,
                    source_content=(
                        [row["source_text"]] if "source_text" in row else []
                    ),
                )
                for row in reader
            ]
        except KeyError:
            raise click.UsageError(
                "CSV file must have columns 'question' and 'expected_answer'. A header is required. An optional source_text column can be included."
            )

    dataset_id = client.upload_dataset(name, dataset)
    print(f"Dataset created with ID: {dataset_id}")


@evals.command()
@click.option(
    "--api_key",
    default=talc_api_key,
    help="Talc API key. Can also be specified with the TALC_API_KEY environment variable.",
)
@click.option(
    "--in_file",
    prompt="Input CSV path",
    help="Path to CSV file to upload. Columns should be 'question', 'expected_answer' and 'user_answer'. An optional source_text column can be included as well.",
)
@click.option(
    "--name",
    help="The name of the dataset.",
    default="Temp Dataset",
    required=False,
)
@click.option(
    "--out_file",
    help="Name of the output file",
    required=False,
)
@click.option("--print_passed", is_flag=True, help="Print the passed test cases.")
def upload_and_grade(
    api_key: str | None,
    in_file: str,
    name: str,
    out_file: str | None,
    print_passed: bool,
):
    """Upload a CSV with questions, reference answers, and user answers. Grade the user answers and return the grade. Column names must be 'question', 'expected_answer', and 'user_answer'."""

    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    client = TalcClient(api_key, talc_base_url)

    with open(in_file, newline="", encoding="utf-8", errors="replace") as csvfile:
        reader = csvmod.DictReader(csvfile)

        rows = [row for row in reader]

        assert "question" in rows[0]
        assert "expected_answer" in rows[0]
        assert "user_answer" in rows[0]

    dataset = [
        TestCaseWithRubric(
            question=row["question"],
            expected_response=row["expected_answer"],
            scenario_data={},
            id=None,
            source_content=([row["source_text"]] if "source_text" in row else []),
        )
        for row in rows
    ]

    dataset = client.upload_dataset(name, dataset)

    # Download the dataset to get the IDs
    data = client.get_dataset(dataset.id)

    # Map the IDs to the actual responses using the questions as a key

    id_map = {case.question: case.id for case in data}

    results = [
        TestCaseResponse(
            id=id_map[row["question"]],  # type: ignore
            response=row["user_answer"] if "user_answer" in row else "",
        )
        for row in rows
    ]

    # Create a run and submit the results
    run_info = client.start_run()

    # MAX TODO: Expose this config to the user
    grading_set = GradingSet(
        responses=results,
        pipeline=DEFAULT_GRADER_CONFIG,
    )
    client.submit_responses(grading_set)
    print("Submitted results. Beginning grading for run ID: " + run_info.id)

    res = poll_results(
        client, print_passed, out_file, DEFAULT_GRADER_CONFIG.scoring_rules
    )

    exit(res)


def poll_generation_job(
    client: TalcClient, dataset_job: DatasetCreationJob, suppress: bool
):
    lines_to_clear = 0

    def print_status():
        CLEAR_LINE = "\33[2K"
        MOVE_UP = "\033[A"
        MOVE_START = "\r"

        nonlocal lines_to_clear

        print((MOVE_UP + MOVE_START + CLEAR_LINE) * lines_to_clear, end="")

        to_output = ""
        to_output += (
            f"Dataset ID: {dataset_job.dataset_id}\n"
            f"Status: {dataset_job.status} (last updated {dataset_job.last_updated})\n"
        )
        if dataset_job.diagnostics is not None:
            if "kbs" in dataset_job.diagnostics:
                to_output += "KBs:\n"
                to_output += "".join(
                    [indent(kb, "\t") + "\n" for kb in dataset_job.diagnostics["kbs"]]
                )
            if "questions" in dataset_job.diagnostics:
                to_output += "Configs:\n"
                to_output += "".join(
                    [
                        indent(question, "\t") + "\n"
                        for question in dataset_job.diagnostics["questions"]
                    ]
                )

        print(to_output)
        lines_to_clear = to_output.count("\n") + 1

    while True:
        dataset_job = client.get_dataset_creation_job(dataset_job.id)
        not suppress and print_status()
        if dataset_job.status == "COMPLETE":
            break
        if dataset_job.status == "ERROR":
            print("Generation failed.")
            print(dataset_job.error)
            exit(1)
        time.sleep(2)


@evals.command()
@click.option(
    "--api_key",
    default=talc_api_key,
    help="Talc API key. Can also be specified with the TALC_API_KEY environment variable.",
)
@click.option("--out_file", default=None, help="Path to save the output to.")
@click.option(
    "--config",
    help="Path to a JSON config file.",
    default=None,
)
@click.option(
    "--question_limit",
    default=10,
    type=click.INT,
    help="The number of questions to generate. If set in the config file, this will be ignored.",
)
@click.option(
    "--source",
    help="Input folder, single file path, or URL. Multiple allowed.",
    multiple=True,
    default=[],
)
@click.option(
    "--local_save",
    help="Should the knowledge base be saved locally?",
    default=True,
)
@click.option(
    "--suppress",
    help="Suppresses all terminal output besides dataset UUID.",
    default=False,
    is_flag=True,
)
def generate(
    api_key: str | None,
    config: str | None,
    out_file: str | None,
    question_limit: int,
    source: list[str],
    local_save: bool,
    suppress: bool,
):
    """Generate a dataset from a set of documents."""
    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    if out_file is None:
        out_file = (
            "questions_"
            + datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
            + ".csv"
        )

    client = TalcClient(api_key, talc_base_url)

    kb_override_id = None

    if len(source) > 0:
        documents = parse_documents(source)
        knowledge_base = client.upload_knowledge_base_asbytes(
            documents,
            "questions_" + datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S"),
        )
        not suppress and print(f"Knowledge Base created with ID: {knowledge_base.id}")
        kb_override_id = knowledge_base.id

    if config is None:
        generation_configs = []

        s1 = DEFAULT_SIMPLE_ARGS.copy()
        s1.question_limit = ceil(question_limit / 3)
        generation_configs.append(s1)
        s2 = DEFAULT_MID_SIMPLE_ARGS.copy()
        s2.question_limit = ceil(question_limit / 3)
        generation_configs.append(s2)
        s3 = DEFAULT_HALLUCINATION_ARGS.copy()
        s3.question_limit = ceil(question_limit / 3)
        generation_configs.append(s3)

        if kb_override_id is None:
            raise click.UsageError(
                "No source documents provided. Please use the --source option or provide a config file."
            )

        generation_args = QuestionGenerationConfig(
            knowledge_bases=[kb_override_id],
            question_configs=generation_configs,
            name="questions_" + datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S"),
        )
    else:
        try:
            generation_args = QuestionGenerationConfig.parse_file(config)
        except Exception as e:
            raise click.UsageError(f"Error parsing config file: {e}")
        if kb_override_id is not None:
            generation_args.knowledge_bases = [kb_override_id]
        elif len(generation_args.knowledge_bases) == 0:
            raise click.UsageError("No knowledge bases found in the config file.")
        if len(generation_args.question_configs) == 0:
            raise click.UsageError(
                "No question configurations found in the config file."
            )

    dataset_job = client.start_generate_dataset(
        config=generation_args,
    )

    poll_generation_job(client, dataset_job, suppress)

    not suppress and print(f"Dataset created with ID:")
    print(dataset_job.dataset_id)

    if not local_save:
        return

    dataset = client.get_dataset(dataset_job.dataset_id)
    with open(out_file, "w", newline="", encoding="utf-8", errors="replace") as csvfile:
        writer = csvmod.writer(csvfile)
        writer.writerow(["id", "question", "expected_answer", "sources", "type"])
        for case in dataset:
            generated_by: str = str(
                case.scenario_data.get("source_generator", "")
            ).removesuffix("QuestionGenerator")
            writer.writerow(
                [
                    case.id,
                    case.question,
                    case.expected_response,
                    case.scenario_data.get("source_document_titles", []),
                    generated_by,
                ]
            )

    not suppress and print(f"Dataset written to {out_file}")


@evals.command()
@click.option(
    "--api_key",
    default=talc_api_key,
    help="Talc API key. Can also be specified with the TALC_API_KEY environment variable.",
)
@click.option("--dataset", prompt="Dataset ID", help="The ID of the dataset to load.")
@click.option(
    "--out_file",
    help="Path to CSV file to save results to.",
)
@click.option(
    "--wait",
    is_flag=True,
    help="When set, waits for the dataset to finish generating and then downloads it.",
)
def get_dataset(dataset: str, api_key: str | None, out_file: str, wait: bool) -> None:
    """Download or pretty print a test dataset from the server."""

    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    client = TalcClient(api_key, talc_base_url)

    # If the dataset generation is not complete, print the status and exit
    dataset_job = client.get_dataset_status(dataset)
    if dataset_job.status != "COMPLETE":
        if wait:
            poll_generation_job(client, dataset_job, False)
        else:
            to_output = ""
            to_output += (
                f"Dataset ID: {dataset_job.dataset_id}\n"
                f"Status: {dataset_job.status} (last updated {dataset_job.last_updated})\n"
            )
            if dataset_job.diagnostics is not None:
                if "kbs" in dataset_job.diagnostics:
                    to_output += "KBs:\n"
                    to_output += "".join(
                        [
                            indent(kb, "\t") + "\n"
                            for kb in dataset_job.diagnostics["kbs"]
                        ]
                    )
                if "questions" in dataset_job.diagnostics:
                    to_output += "Configs:\n"
                    to_output += "".join(
                        [
                            indent(question, "\t") + "\n"
                            for question in dataset_job.diagnostics["questions"]
                        ]
                    )

            print(to_output)
            exit(0)

    data = client.get_dataset(dataset)
    with open(out_file, "w", newline="", encoding="utf-8", errors="replace") as csvfile:
        writer = csvmod.writer(csvfile)
        writer.writerow(["id", "question", "expected_answer", "sources", "type"])
        for case in data:
            generated_by: str = str(
                case.scenario_data.get("source_generator", "")
            ).removesuffix("QuestionGenerator")

            writer.writerow(
                [
                    case.id,
                    case.question,
                    case.expected_response,
                    case.scenario_data.get("source_document_titles", []),
                    generated_by,
                ]
            )


@evals.command
@click.option(
    "--api_key",
    default=talc_api_key,
    help="Talc API key. Can also be specified with the TALC_API_KEY environment variable.",
)
def list_datasets(api_key: str | None):
    """List the most recent datasets along with their status. Some of these may still be in the generation process."""
    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )
    client = TalcClient(api_key, talc_base_url)
    datasets_jobs = client.get_recent_datasets()

    print("ID - Name - Status")
    for dataset in datasets_jobs:
        print(f"{dataset.dataset_id}\t|\t{dataset.name}\t|\t{dataset.status}")


def parse_documents(input_path: list[str]) -> list[Document]:
    documents: list[Document] = []
    web = Web()
    # Figure out what kind of document each file is, and load it into a list of Document objects
    for f in input_path:
        # If it's a URL, load it locally and scrape
        if f.startswith("http") or f.startswith("https"):
            scraped = web.scrape(f)
            if len(scraped) > 0:
                for doc in scraped:
                    documents.append(doc)
        elif os.path.isdir(f):
            for root, _, files in os.walk(f):
                for path in files:
                    document = read_document(os.path.join(root, path))
                    if document is not None:
                        documents.append(document)
                    else:
                        # TODO: [Resolve handling] Current behavior is to gracefully skip if unsupported file within directory, but we throw error if unsupported file is provided directly as only input.
                        print(
                            f"Skipping file with unknown extension {path} in directory {root}."
                        )
        elif os.path.isfile(f):
            document = read_document(f)
            if document is not None:
                documents.append(document)
            else:
                raise click.UsageError(
                    "Unsupported file type. Only .md and .pdf files are supported."
                )
        else:
            raise click.UsageError(
                "File not found or invalid URL. Please provide a valid file path or URL."
            )
    return documents


def read_document(fp: str) -> Document | None:
    """
    Read a document from a file path as either text or bytes.
    Currently supported extensions: [.pdf, .md, .txt]
    """
    if fp.endswith(".md") or fp.endswith(".txt"):
        with open(fp, "r") as f:
            return Document(
                content=f.read().encode(encoding="utf-8", errors="ignore"),
                title=f.name,
                filepath=os.path.basename(fp),
                content_type="text/markdown",
            )
    if fp.endswith(".txt"):
        with open(fp, "r") as f:
            return Document(
                content=f.read().encode(encoding="utf-8", errors="ignore"),
                title=f.name,
                filepath=os.path.basename(fp),
                content_type="text/plain",
            )
    elif fp.endswith(".pdf"):
        print(f"Reading pdf: {fp}")
        with open(fp, "rb") as f:
            return Document(
                content=f.read(),
                title=os.path.basename(fp),
                filepath=fp,
                content_type="application/pdf",
            )
    else:
        return None


@evals.command()
@click.option(
    "--api_key",
    default=talc_api_key,
    help="Talc API key. Can also be specified with the TALC_API_KEY environment variable.",
)
@click.option(
    "--source",
    help="Input folder, single file path, or URL. Multiple allowed.",
    multiple=True,
    default=[],
)
@click.option(
    "--source_list",
    default=None,
    help="A list of sources to upload. Accepts a newline separated txt file of files, filepaths, or URLs to upload.",
)
@click.option(
    "--friendly_name",
    default=None,
    help="Brief text description of this knowledge base.",
)
@click.option(
    "--verify",
    is_flag=True,
    default=False,
    help="Verifies the uploaded knowledge base to check for potential errors; this is currently experimental.",
)
@click.option(
    "--auto_link",
    is_flag=True,
    default=False,
    help="Identifies documents that reference each other and links them; this is currently experimental.",
)
def upload_kb(
    api_key: str | None,
    source: list[str],
    friendly_name: str | None,
    source_list: str | None,
    verify: bool,
    auto_link: bool,
):
    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    # if knowledge_base_name is None:
    #    knowledge_base_name = f"kb_" + datetime.datetime.now().strftime(
    #        "%Y_%m_%d_%H_%M_%S"
    #    )
    if friendly_name is None:
        raise click.UsageError(
            "No friendly name  provided. Please use the --friendly_name option to provide a brief description string"
        )

    if source_list is not None:
        with open(source_list, "r") as f:
            source = f.read().split("\n")
            source = [x for x in source if x != ""]

    client = TalcClient(api_key, talc_base_url)
    documents = parse_documents(source)
    print("Uploading documents to create a knowledge base. This may take a while...")
    knowledge_base = client.upload_knowledge_base_asbytes(
        documents, friendly_name, verify=verify, auto_link=auto_link
    )

    if verify and knowledge_base.errors is not None:
        print()
        if len(knowledge_base.errors) == 0:
            print(
                colored(
                    "No errors were found while verifying the uploaded knowledge base.",
                    "green",
                )
            )
        else:
            print(
                colored(
                    "Potential errors were found while verifying the uploaded knowledge base:",
                    "yellow",
                )
            )
            for error in knowledge_base.errors:
                print(colored(f"  • {error}", "yellow"))
        print()

    print(
        colored(
            f"Knowledge Base created with ID: {colored(knowledge_base.id, attrs=['bold'])}",
            "green",
        )
    )


@evals.command()
@click.option(
    "--api_key",
    default=talc_api_key,
    help="Talc API key. Can also be specified with the TALC_API_KEY environment variable.",
)
def list_kbs(
    api_key: str | None,
):
    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )
    client = TalcClient(api_key, talc_base_url)
    kbs = client.list_kbs()
    print("Kbs:")
    for kb in kbs:
        print(f"  {kb.friendly_name} ({kb.id})")


def main():
    evals()



from pydantic import BaseModel, Field, SerializeAsAny  # type: ignore
from typing import Literal, Optional, Union, Annotated


class Document(BaseModel):
    """Represents a document used as a ground truth when generating tests."""

    # url: Optional[str] = None
    filepath: str
    title: Optional[str] = None
    content: bytes | str
    content_type: Optional[str] = "application/octet-stream"


class QuestionWithReason(BaseModel):
    reason: str = Field(
        description="Used for explaining the reasoning behind the question.",
        default="<Placeholder for reasoning>",
    )
    question: str = Field(description="The question text.")
    reference_answer: str = Field(description="The expected response.")


class QuestionGeneratorArgs(BaseModel):
    question_limit: int = Field(description="The number of questions to generate")


class SimpleQuestionGeneratorArgs(QuestionGeneratorArgs):
    question_type: Literal["SimpleQuestion"] = (
        "SimpleQuestion"  # Type discriminator used for deserialization. Do not change.
    )
    temperature: float | None = Field(
        description="The temperature to use when generating questions", default=None
    )
    positive_examples: list[QuestionWithReason] = Field(
        description="A list of few-shot examples of good questions to help guide the question generation"
    )
    negative_examples: list[QuestionWithReason] = Field(
        description="A list of few-shot examples of poorly phrased or incorrect questions to help guide the question generation",
        default=[],
    )
    extra_criteria: list[str] = Field(
        description="A list of extra criteria to evaluate the quality of the generated questions.",
        default=[],
    )


class HallucinationInducingQuestionGeneratorArgs(QuestionGeneratorArgs):
    question_type: Literal["HallucinationInducing"] = (
        "HallucinationInducing"  # Type xdiscriminator used for deserialization. Do not change.
    )
    temperature: float | None = Field(
        description="The temperature to use when generating questions", default=None
    )
    positive_examples: list[QuestionWithReason] = Field(
        description="A list of few-shot examples of good questions to help guide the question generation"
    )
    negative_examples: list[QuestionWithReason] = Field(
        description="A list of few-shot examples of poorly phrased or incorrect questions to help guide the question generation",
        default=[],
    )
    extra_criteria: list[str] = Field(
        description="A list of extra criteria to evaluate the quality of the generated questions.",
        default=[],
    )


class LinkedTopicQuestionGeneratorArgs(QuestionGeneratorArgs):
    question_type: Literal["LinkedTopic"] = (
        "LinkedTopic"  # Type xdiscriminator used for deserialization. Do not change.
    )
    temperature: float | None = Field(
        description="The temperature to use when generating questions", default=None
    )
    positive_examples: list[QuestionWithReason] = Field(
        description="A list of few-shot examples of good questions to help guide the question generation"
    )
    negative_examples: list[QuestionWithReason] = Field(
        description="A list of few-shot examples of poorly phrased or incorrect questions to help guide the question generation",
        default=[],
    )
    extra_criteria: list[str] = Field(
        description="A list of extra criteria to evaluate the quality of the generated questions.",
        default=[],
    )


# This is a discriminated union. What we want is to have a list of question generator configs, and to be able to send that list from the client/cli to the server and have it correctly deserialize.
# Breaking it down:
#  - Union means the type GeneratorArgs can refer to EITHER SimpleQuestionGeneratorArgs or HallucinationInducingQuestionGeneratorArgs
#  - When deserializing one of these GeneratorArgs over the wire, look at the question_type field to determine which of the two we're currently deserializing. It's a constant with a different value for the two classes (SimpleQuestion or HallucinationInducing)
#  - The SerializeAsAny is just there to suppress some warnings caused by a bug in pydantic.
GeneratorArgs = SerializeAsAny[
    Annotated[
        Union[
            SimpleQuestionGeneratorArgs,
            HallucinationInducingQuestionGeneratorArgs,
            LinkedTopicQuestionGeneratorArgs,
        ],
        Field(discriminator="question_type"),
    ]
]


class QuestionGenerationConfig(BaseModel):
    name: str = Field(
        description="Friendly name for the configuration used in logging, error reporting and the UI."
    )
    knowledge_bases: list[str] = Field(
        description="List of knowledge bases to use for generating questions. Uses knowledge base UUIDs."
    )
    question_configs: list[GeneratorArgs] = Field(  # type: ignore
        description="List of question generator configurations."
    )



from typing import Literal
from pydantic import BaseModel, ConfigDict, Extra, SerializeAsAny  # type: ignore


class TestCaseResponse(BaseModel):
    """Represents the results of running a single test case."""

    id: str
    response: str


class GraderConfig(BaseModel):
    """The basic configuration used to run a grader. Most graders will have additional configuration options derived from this class."""

    grader: str
    config_name: str
    description: str

    model_config = ConfigDict(
        extra="allow",  # type: ignore
    )


class ScoringRule(BaseModel):
    """Represents the rules used to take a list of grades and produce a final score."""

    graders: list[str]
    level: Literal["warn", "fail"]
    mode: Literal["allow_any_pass", "require_all_pass"]


class GradingPipeline(BaseModel):
    """Configures a pipeline of graders that should be run on a test case."""

    graders: list[SerializeAsAny[GraderConfig]]
    scoring_rules: list[ScoringRule]


class FewShotConfig(BaseModel):
    reason: str
    question: str
    correct_answer: str | None
    user_answer: str
    grade: str


class FactualityGraderConfig(GraderConfig):
    """Configuration for the factuality grader."""

    few_shot_examples: list[FewShotConfig]
    additional_pass_criteria: list[str]
    additional_fail_criteria: list[str]


class FailIfGraderConfig(GraderConfig):
    """Configuration for the grader that checks for explicitly disallowed cases. Passes if the response does not match any of the fail_criteria, fails if it does."""

    fail_criteria: list[str]
    """The criteria that, if met, will cause the grader to fail the test case. e.g, 'If the response contains 'as an AI language model'.'"""


class GradingSet(BaseModel):
    """Represents a set of test cases that should be graded, and the methods for grading them."""

    responses: list[TestCaseResponse]
    pipeline: GradingPipeline



from talc.synthetic import (
    QuestionWithReason,
    HallucinationInducingQuestionGeneratorArgs,
    SimpleQuestionGeneratorArgs,
    QuestionGenerationConfig,
)

DEFAULT_HALLUCINATION_ARGS = HallucinationInducingQuestionGeneratorArgs(
    question_limit=10,
    positive_examples=[
        QuestionWithReason(
            reason="The documentation describes ways to get the maximum and minimum values of a numpy array, but does not describe how to get the median value.",
            question="What numpy method of numpy.ndarray objects returns the indices of the median value along an axis?",
            reference_answer="There is no numpy.ndarray method that returns the indices of the median value along an axis.",
        ),
        QuestionWithReason(
            reason="The article describes the history of John D Rockefeller Sr. He died in 1937, so we can ask a question about something he did after that year in order to induce hallucination.",
            question="What job prompted John D Rockefeller Sr to move to Brazil in 1993?",
            reference_answer="John D Rockefeller Sr did not move to Brazil in 1993. He died in 1937.",
        ),
    ],
)

DEFAULT_SIMPLE_ARGS = SimpleQuestionGeneratorArgs(
    question_limit=10,
    positive_examples=[
        QuestionWithReason(
            question="Dippen Bay is an embayment along Kilbrannan Sound on the east coast of what peninsula?",
            reference_answer="Kintyre Peninsula",
            reason="The question is unambiguous and has a clear answer.",
        ),
        QuestionWithReason(
            question="Who currently owns the Bentley S1 Drophead Coupe orignally owned by John D Rockefeller Jr?",
            reference_answer="Henrik Frederiksen",
            reason="The question is specific and requires knowledge of a specific fact. It's not a generic question that could apply to any page, and has an unambiguous answer.",
        ),
    ],
)

DEFAULT_MID_SIMPLE_ARGS = SimpleQuestionGeneratorArgs(
    question_limit=10,
    positive_examples=[
        QuestionWithReason(
            question="A railway staton was opened by the Belfast and County Down Railway on June 3 1861. What was the original name of the station?",
            reference_answer="Groomsport and Bangor",
            reason="By avoiding using the current name of the station (but still providing enough context to identify the station), we can ask a question that is slightly more difficult than a simple question.",
        ),
        QuestionWithReason(
            question="In 1890, Grover Cleveland purchased a summer home. What event caused him to stop using the home?",
            reference_answer="The death of his daughter Ruth",
            reason="By refering to the topic indirectly (a summer home purchased in 1890, rather than the name of the home), we can make answering the question more difficult for models that are using retrieval-based methods.",
        ),
    ],
    extra_criteria=[
        "If there is a title of the page, avoid using it directly in the question. Instead, refer to the topic indirectly using other facts to specify it.",
    ],
)



import os
import csv
from textwrap import indent, wrap
from typing import Literal
from termcolor import colored
from talc.talc_client import (
    GradedTestCase,
)
from talc.synthetic import Document
from talc.grading import ScoringRule
import requests
import aiohttp
from readabilipy import simple_json_from_html_string
from markdownify import markdownify as md


class Web:
    def __get_readable(self, url, html: str, use_readability: bool = True):
        """Parse HTML and return the knowledge base as a list of documents. Only used when as_bytes=False."""
        readable = simple_json_from_html_string(html, use_readability=use_readability)
        content = md(readable["plain_content"])
        document = Document(
            title=url,
            content=content,
            filepath=url,
            content_type="text/markdown",
        )
        return [document]

    def scrape(self, url: str, use_readability: bool = True) -> list[Document]:
        """Synchronously scrape a webpage and return a list of Documents."""
        # Note: Moving linking and chunking to later stages after kb upload.

        print("Scraping doc: " + url)
        response = requests.get(url)

        text: str = response.text

        raw_bytes: bytes = response.content
        if response.headers.get("Content-Type") == "text/html":
            title = simple_json_from_html_string(text, use_readability=use_readability)[
                "title"
            ]
        else:
            title = url
        document = Document(
            title=url,
            content=raw_bytes,
            filepath=url,
            content_type=response.headers.get(
                "Content-Type"
            ),  # Note: supabase storage saves text/html as text/plain for security reasons
        )
        return [document]

    async def ascrape(
        self, url: str, use_readability: bool = True, as_bytes=True
    ) -> list[Document]:
        """Asynchronously scrape a webpage and return a list of Documents."""
        # Note: Moving linking and chunking to later stages after kb upload.

        print("Scraping doc: " + url)

        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                response.raise_for_status()

                text: str = await response.text()
                if as_bytes:
                    raw_bytes: bytes = await response.read()
                    title = simple_json_from_html_string(
                        text, use_readability=use_readability
                    )["title"]
                    document = Document(
                        title=url,
                        content=raw_bytes,
                        filepath=url,
                        content_type=response.headers.get(
                            "Content-Type"
                        ),  # Note: supabase storage saves text/html as text/plain for security reasons
                    )
                    return [document]
                else:
                    return self.__get_readable(
                        url, text, use_readability=use_readability
                    )


# Ugly terminal output functions
def print_single_case(case: GradedTestCase, color: Literal["red", "yellow", "green"]):

    try:
        terminal_width = max(60, os.get_terminal_size().columns)
    except OSError:
        terminal_width = 80

    response_wrapped = "\n".join(
        wrap(f"Response: {case.response}", terminal_width - 10)
    )
    expected_wrapped = "\n".join(
        wrap(f"Expected: {case.expected_response}", terminal_width - 10)
    )
    question_wrapped = "\n".join(
        wrap(f"{case.id}: {case.question}\n", terminal_width - 10)
    )

    print(colored(question_wrapped, color, attrs=["bold"]))
    print(indent(response_wrapped, "\t"))
    print(indent(expected_wrapped, "\t"))
    for grade in case.grades:
        if grade.score >= 1:
            continue
        else:
            grader_color = color
        grade_wrapped = "\n".join(
            wrap(
                f"{grade.grader} {'PASS' if grade.score >= 1.0 else 'FAIL'}: {grade.reason}",
                terminal_width - 10,
            )
        )
        print(colored(grade_wrapped, grader_color))
    print("---------------------------------------------")


def get_failures_and_warnings(
    test_cases: list[GradedTestCase], scoring_rules: list[ScoringRule]
):

    failures = []
    warnings = []

    # Now we need to loop through the rules defined in the scoring rules and
    # determine which cases need to be considered failures vs warnings.
    # This is ugly because it needs to support "any" and "all" scoring.
    for test_case in test_cases:
        for scoring_rule in scoring_rules:
            # determine if the test case has grades for all of the graders for this rule
            eligible = all(
                [
                    any(map(lambda x: x.grader == grader, test_case.grades))
                    for grader in scoring_rule.graders
                ]
            )
            if eligible and scoring_rule.mode == "allow_any_pass":
                all_fail = all(
                    [
                        all(
                            map(
                                lambda x: x.grader != grader or x.score < 1,
                                test_case.grades,
                            )
                        )
                        for grader in scoring_rule.graders
                    ]
                )
                if all_fail:
                    if scoring_rule.level == "fail":
                        failures.append(test_case)
                    elif scoring_rule.level == "warn":
                        warnings.append(test_case)
            elif eligible and scoring_rule.mode == "require_all_pass":
                any_fail = any(
                    [
                        any(
                            map(
                                lambda x: x.grader == grader and x.score < 1,
                                test_case.grades,
                            )
                        )
                        for grader in scoring_rule.graders
                    ]
                )
                if any_fail:
                    if scoring_rule.level == "fail":
                        failures.append(test_case)
                    elif scoring_rule.level == "warn":
                        warnings.append(test_case)

    return failures, warnings



"""
Some experimental patterns for simplifying using the CLI for testing; 
may or may not be helpful for real-world use cases.
"""

import pathlib
import os
import re
import shutil
import subprocess
import click
from pydantic import BaseModel
from pydantic_yaml import parse_yaml_file_as, to_yaml_file
from talc.synthetic import (
    GeneratorArgs,
    QuestionGenerationConfig,
    QuestionGeneratorArgs,
)

UUID_PATTERN = re.compile(
    "[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}"
)

TALCUP_HOME = pathlib.Path.home() / "talc"

class LocalConfigSourceList(BaseModel):
    name: str
    sources: list[str] | None = None
    source_lists: list[str] | None = None


class LocalConfigRun(BaseModel):
    name: str
    questions: list[GeneratorArgs]
    sources: list[str] | None = None
    source_lists: list[str] | None = None


class LocalConfigFile(BaseModel):
    source_lists: list[LocalConfigSourceList] | None = None
    runs: list[LocalConfigRun] | None = None


class CachedKnowledgeBase(BaseModel):
    id: str
    sources: list[str]


class LocalConfig:
    source_lists: dict[str, LocalConfigSourceList]
    runs: dict[str, LocalConfigRun]

    def __init__(self):
        self.source_lists: dict[str, LocalConfigSourceList] = {}
        self.runs: dict[str, LocalConfigRun] = {}

    def add_source_list(self, list: LocalConfigSourceList):
        if list.name in self.source_lists:
            raise Exception(f"source list already exists: '{list.name}'")
        self.source_lists[list.name] = list

    def add_run(self, run: LocalConfigRun):
        if run.name in self.runs:
            raise Exception(f"run already exists: '{run.name}'")
        self.runs[run.name] = run

    def resolve_run(self, run_name: str) -> tuple[set[str], list[GeneratorArgs]]:
        if run_name not in self.runs:
            raise Exception(f"run does not exist: '{run_name}'")
        run = self.runs[run_name]
        sources = set()
        if run.sources is not None:
            for source in run.sources:
                sources.add(source)
        if run.source_lists is not None:
            for source_list in run.source_lists:
                for source in self.resolve_source_list(source_list):
                    sources.add(source)
        return sources, run.questions

    def resolve_source_list(self, source_list_name: str) -> set[str]:
        if source_list_name not in self.source_lists:
            raise Exception(f"source list does not exist: '{source_list_name}'")
        source_list = self.source_lists[source_list_name]
        sources = set()
        if source_list.sources is not None:
            for source in source_list.sources:
                sources.add(source)
        if source_list.source_lists is not None:
            for other_source_list_name in source_list.source_lists:
                for source in self.resolve_source_list(other_source_list_name):
                    sources.add(source)
        return sources


class LocalConfigLoader:

    def __init__(self):
        self.config_dir = TALCUP_HOME

    def _ensure_config_dir(self):
        if not self.config_dir.exists():
            self.config_dir.mkdir()

    def load(self) -> LocalConfig:
        self._ensure_config_dir()

        config = LocalConfig()
        for file in os.listdir(self.config_dir):
            if not file.endswith(".yaml"):
                continue
            parsed = parse_yaml_file_as(LocalConfigFile, self.config_dir / file)
            if parsed.source_lists:
                for source_list in parsed.source_lists:
                    config.add_source_list(source_list)
            if parsed.runs:
                for run in parsed.runs:
                    config.add_run(run)
        return config


class LocalConfigCacheLoader:

    def __init__(self):
        self.cache_dir = TALCUP_HOME / ".kb_cache"

    def _ensure_cache_dir(self):
        if not self.cache_dir.exists():
            self.cache_dir.mkdir()

    def add_cached_kb(self, kb_id: str, sources: set[str]):
        self._ensure_cache_dir()
        cached = CachedKnowledgeBase(id=kb_id, sources=list(sources))
        cached.id = kb_id
        cached.sources = list(sources)
        to_yaml_file(self.cache_dir / f"{kb_id}.yaml", cached)

    def find_cached_kbs(self, sources: set[str]) -> list[str]:
        self._ensure_cache_dir()
        kb_ids = []
        for file in os.listdir(self.cache_dir):
            if not file.endswith(".yaml"):
                continue
            parsed = parse_yaml_file_as(CachedKnowledgeBase, self.cache_dir / file)
            if set(parsed.sources) == sources:
                kb_ids.append(parsed.id)
        return kb_ids


@click.group()
def talcup():
    pass


@talcup.command(name="init")
def init_example():
    example_src_path = (
        pathlib.Path(__file__).resolve().parent / "talcup_example.yaml"
    ).resolve()
    example_dest_path = TALCUP_HOME / "example.yaml"
    if example_dest_path.exists():
        print(f"Can't create example at {example_dest_path}: a file already exists.")
        return
    shutil.copy(example_src_path, example_dest_path)
    print(f"Created an example config file at {example_dest_path}.")
    print("Try running:")
    print("  talcup list")
    print("  talcup show examplerun")
    print("  talcup gen examplerun")


def print_empty_message():
    print(f"No runs are defined. Add config files to {TALCUP_HOME}.")
    print("You can also run `talcup init` to create an example config file.")


@talcup.command(name="list")
def list_runs():
    config = LocalConfigLoader().load()
    if not config.runs or len(config.runs) == 0:
        print_empty_message()
        return

    run_names = sorted(config.runs)

    print(f"Available runs (from {TALCUP_HOME}):")
    for run in config.runs:
        meta: str
        try:
            sources, questions = config.resolve_run(run)
            meta = f"({len(sources)} sources, {len(questions)} question configs)"
        except:
            meta = "(couldn't be fully resolved due to an error)"
        print(f" - {run} {meta}")


@talcup.command
@click.argument("name")
def show(name: str):
    config = LocalConfigLoader().load()
    if not config.runs or len(config.runs) == 0:
        print_empty_message()
        return

    sources, questions = config.resolve_run(name)
    print("Sources:")
    for source in sources:
        print(f" - {source}")
    print("Questions:")
    for question in questions:
        print(f" - {question.question_type}")


@talcup.command
@click.option("--force-upload", is_flag=True)
@click.argument("name")
def gen(name: str, force_upload: bool):
    config = LocalConfigLoader().load()
    if not config.runs or len(config.runs) == 0:
        print_empty_message()
        return

    sources, questions = config.resolve_run(name)

    cache_loader = LocalConfigCacheLoader()
    maybe_kbs = cache_loader.find_cached_kbs(sources)
    ls_completed = subprocess.run(["talc", "list-kbs"], capture_output=True)
    actual_kbs = str(ls_completed.stdout, "utf-8")
    matching_kbs = [kb for kb in maybe_kbs if kb in actual_kbs]

    if len(matching_kbs) == 0 or force_upload:
        if force_upload:
            print("Forcing a new KB upload...")
        else:
            print("Couldn't find a cached KB; uploading one now...")
        args = ["talc", "upload-kb"]
        for source in sources:
            args.append("--source")
            args.append(source)
        args.append("--friendly_name")
        args.append(name)
        upload_completed = subprocess.run(args, capture_output=True)
        uploaded_kb_ids = UUID_PATTERN.findall(str(upload_completed.stdout, "utf-8"))
        if len(uploaded_kb_ids) != 1:
            raise Exception("expected upload-kb to return a single UUID")
        kb_id = uploaded_kb_ids[0]
        cache_loader.add_cached_kb(kb_id, sources)
        print(f"  Added KB: {kb_id}")
    else:
        kb_id = matching_kbs[0]
        print(f"Found cached KB: {kb_id}")

    parent_output_dir = TALCUP_HOME / "output"
    if not parent_output_dir.exists():
        parent_output_dir.mkdir()

    output_dir = parent_output_dir / name
    if not output_dir.exists():
        output_dir.mkdir()

    config = QuestionGenerationConfig(
        name=name, knowledge_bases=[kb_id], question_configs=questions
    )
    config_path = output_dir / "config.json"
    with open(config_path, "w") as f:
        f.write(config.model_dump_json(indent=4))

    print("\nRunning talc generate...\n")
    subprocess.run(
        ["talc", "generate", "--config", config_path], cwd=config_path.parent
    )


def main():
    talcup()



from talc.grading import (
    GradingSet,
    GraderConfig,
    FactualityGraderConfig,
    FewShotConfig,
    FailIfGraderConfig,
    GradingPipeline,
    ScoringRule,
)

DEFAULT_FACTUALITY_GRADER_CONFIG = FactualityGraderConfig(
    grader="FactualityGrader",
    config_name="DefaultFactualityGraderConfig",
    description="The default configuration for the factuality grader.",
    few_shot_examples=[
        FewShotConfig(
            question="What command is used to install the Python SDK for modelscope?",
            correct_answer="The Python SDK for modelscope can be installed using `pip install modelscope` command.",
            user_answer="pip install modelscope",
            grade="PASS",
            reason="The user answer specifies the correct command.",
        ),
    ],
    additional_pass_criteria=[
        "It's ok if the user response includes extra information as long as it does not contradict the reference answer.",
        "If the expected answer is 'I don't know' or similar, then any answer that expresses a definite answer should be considered a failing grade.",
    ],
    additional_fail_criteria=[],
)

DEFAULT_SOURCE_CONTENT_FACTUALITY_GRADER_CONFIG = FactualityGraderConfig(
    grader="SourceContentFactualityGrader",
    config_name="DefaultSourceContentFactualityGraderConfig",
    description="The default configuration for the source content factuality grader.",
    few_shot_examples=[
        FewShotConfig(
            question="What command is used to install the Python SDK for modelscope?",
            correct_answer="The Python SDK for modelscope can be installed using `pip install modelscope` command.",
            user_answer="pip install modelscope",
            grade="PASS",
            reason="The user answer specifies the correct command.",
        ),
    ],
    additional_pass_criteria=[
        "It's ok if the user response includes extra information as long as it does not contradict the source contents.",
    ],
    additional_fail_criteria=[],
)

DEFUALT_FAIL_IF_GRADER_CONFIG = FailIfGraderConfig(
    grader="FailIfGrader",
    config_name="DefaultFailIfGraderConfig",
    description="The default configuration for the fail-if grader.",
    fail_criteria=[
        "The AI responding breaks character by using phrases like 'as an AI agent' or similar.",
    ],
)

DEFAULT_GRADER_CONFIG = GradingPipeline(
    graders=[
        DEFAULT_FACTUALITY_GRADER_CONFIG,
    ],
    scoring_rules=[
        ScoringRule(
            graders=["DefaultFactualityGraderConfig"],
            level="fail",
            mode="require_all_pass",
        )
    ],
)
